"""A collection of modules to perform post-hoc analysis of data.

"""

from osl_dynamics.analysis import connectivity
from osl_dynamics.analysis import power
from osl_dynamics.analysis import spectral
from osl_dynamics.analysis import modes
from osl_dynamics.analysis import time_series
from osl_dynamics.analysis import workbench
